# -*- coding: utf-8 -*-
"sets the options of the configuration objects Config object itself"
# Copyright (C) 2012-2013 Team tiramisu (see AUTHORS for all contributors)
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# The original `Config` design model is unproudly borrowed from
# the rough pypy's guys: http://codespeak.net/svn/pypy/dist/pypy/config/
# the whole pypy projet is under MIT licence
# ____________________________________________________________

from tiramisu.error import RequirementRecursionError, PropertiesOptionError
from tiramisu.i18n import _


class _const:
    """convenient class that emulates a module
    and builds constants (that is, unique names)"""
    class ConstError(TypeError):
        pass

    def __setattr__(self, name, value):
        if name in self.__dict__:
            raise self.ConstError, _("Can't rebind group ({})").format(name)
        self.__dict__[name] = value

    def __delattr__(self, name):
        if name in self.__dict__:
            raise self.ConstError, _("Can't unbind group ({})").format(name)
        raise ValueError(name)


# ____________________________________________________________
class GroupModule(_const):
    "emulates a module to manage unique group (OptionDescription) names"
    class GroupType(str):
        """allowed normal group (OptionDescription) names
        *normal* means : groups that are not master
        """
        pass

    class DefaultGroupType(GroupType):
        """groups that are default (typically 'default')"""
        pass

    class MasterGroupType(GroupType):
        """allowed normal group (OptionDescription) names
        *master* means : groups that have the 'master' attribute set
        """
        pass
# setting.groups (emulates a module)
groups = GroupModule()


def populate_groups():
    "populates the available groups in the appropriate namespaces"
    groups.master = groups.MasterGroupType('master')
    groups.default = groups.DefaultGroupType('default')
    groups.family = groups.GroupType('family')

# names are in the module now
populate_groups()


# ____________________________________________________________
class OwnerModule(_const):
    """emulates a module to manage unique owner names.

    owners are living in `Config._cfgimpl_value_owners`
    """
    class Owner(str):
        """allowed owner names
        """
        pass

    class DefaultOwner(Owner):
        """groups that are default (typically 'default')"""
        pass
# setting.owners (emulates a module)
owners = OwnerModule()


def populate_owners():
    """populates the available owners in the appropriate namespaces

    - 'user' is the generic is the generic owner.
    - 'default' is the config owner after init time
    """
    setattr(owners, 'default', owners.DefaultOwner('default'))
    setattr(owners, 'user', owners.Owner('user'))

    def add_owner(name):
        """
        :param name: the name of the new owner
        """
        setattr(owners, name, owners.Owner(name))
    setattr(owners, 'add_owner', add_owner)

# names are in the module now
populate_owners()


class MultiTypeModule(_const):
    class MultiType(str):
        pass

    class DefaultMultiType(MultiType):
        pass

    class MasterMultiType(MultiType):
        pass

    class SlaveMultiType(MultiType):
        pass

multitypes = MultiTypeModule()


def populate_multitypes():
    setattr(multitypes, 'default', multitypes.DefaultMultiType('default'))
    setattr(multitypes, 'master', multitypes.MasterMultiType('master'))
    setattr(multitypes, 'slave', multitypes.SlaveMultiType('slave'))

populate_multitypes()


#____________________________________________________________
class Setting(object):
    "``Config()``'s configuration options"
    __slots__ = ('properties', 'permissives', 'owner', 'context')

    def __init__(self, context):
        # properties attribute: the name of a property enables this property
        # key is None for global properties
        self.properties = {None: []}  # ['hidden', 'disabled', 'mandatory', 'frozen', 'validator']}
        # permissive properties
        self.permissives = {}
        # generic owner
        self.owner = owners.user
        self.context = context

    #____________________________________________________________
    # properties methods
    def has_properties(self, opt=None, is_apply_req=True):
        "has properties means the Config's properties attribute is not empty"
        return bool(len(self.get_properties(opt, is_apply_req)))

    def get_properties(self, opt=None, is_apply_req=True):
        if opt is None:
            default = []
        else:
            if is_apply_req:
                apply_requires(opt, self.context)
            default = list(opt._properties)
        return self.properties.get(opt, default)

    def has_property(self, propname, opt=None, is_apply_req=True):
        """has property propname in the Config's properties attribute
        :param property: string wich is the name of the property"""
        return propname in self.get_properties(opt, is_apply_req)

    def enable_property(self, propname):
        "puts property propname in the Config's properties attribute"
        props = self.get_properties()
        if propname not in props:
            props.append(propname)
        self.set_properties(props)

    def disable_property(self, propname):
        "deletes property propname in the Config's properties attribute"
        props = self.get_properties()
        if propname in props:
            props.remove(propname)
        self.set_properties(props)

    def set_properties(self, properties, opt=None):
        """save properties for specified opt
        (never save properties if same has option properties)
        """
        if opt is None:
            self.properties[opt] = properties
        else:
            if opt._properties == properties:
                if opt in self.properties:
                    del(self.properties[opt])
            else:
                self.properties[opt] = properties

    def add_property(self, propname, opt, is_apply_req=True):
        properties = self.get_properties(opt, is_apply_req)
        if not propname in properties:
            properties.append(propname)
            self.set_properties(properties, opt)

    def del_property(self, propname, opt, is_apply_req=True):
        properties = self.get_properties(opt, is_apply_req)
        if propname in properties:
            properties.remove(propname)
            self.set_properties(properties, opt)

    #____________________________________________________________
    def validate_properties(self, opt_or_descr, is_descr, is_write,
                            value=None, force_permissive=False,
                            force_properties=None):
        properties = set(self.get_properties(opt_or_descr))
        #remove this properties, those properties are validate in after
        properties = properties - set(['mandatory', 'frozen'])
        set_properties = self.get_properties()
        if force_properties is not None:
            set_properties.extend(force_properties)
        set_properties = set(set_properties)
        properties = properties & set_properties
        if force_permissive is True or self.has_property('permissive', is_apply_req=False):
            properties = properties - set(self.get_permissive())
        properties = properties - set(self.get_permissive(opt_or_descr))
        properties = list(properties)
        raise_text = _("trying to access"
                       " to an option named: {0} with properties"
                       " {1}")
        if not is_descr:
            if self.context.cfgimpl_get_values().is_mandatory_err(opt_or_descr,
                                                                  value,
                                                                  force_properties=force_properties):
                properties.append('mandatory')
            if is_write and (self.has_property('everything_frozen') or (
                    self.has_property('frozen') and
                    self.has_property('frozen', opt_or_descr,
                                      is_apply_req=False))):
                properties.append('frozen')
                raise_text = _('cannot change the value to {0} for '
                               'option {1} this option is frozen')
        if properties != []:
            raise PropertiesOptionError(raise_text.format(opt_or_descr._name,
                                                          str(properties)),
                                        properties)

    def get_permissive(self, opt=None):
        return self.permissives.get(opt, [])

    def set_permissive(self, permissive, opt=None):
        if not isinstance(permissive, list):
            raise TypeError(_('permissive must be a list'))
        self.permissives[opt] = permissive

    #____________________________________________________________
    def setowner(self, owner):
        ":param owner: sets the default value for owner at the Config level"
        if not isinstance(owner, owners.Owner):
            raise TypeError(_("invalid generic owner {0}").format(str(owner)))
        self.owner = owner

    def getowner(self):
        return self.owner

    #____________________________________________________________
    def read_only(self):
        "convenience method to freeze, hidde and disable"
        self.enable_property('everything_frozen')
        self.enable_property('frozen')  # can be usefull...
        self.disable_property('hidden')
        self.enable_property('disabled')
        self.enable_property('mandatory')
        self.enable_property('validator')
        self.disable_property('permissive')

    def read_write(self):
        "convenience method to freeze, hidde and disable"
        self.disable_property('everything_frozen')
        self.enable_property('frozen')  # can be usefull...
        self.enable_property('hidden')
        self.enable_property('disabled')
        self.disable_property('mandatory')
        self.enable_property('validator')
        self.disable_property('permissive')


def apply_requires(opt, config):
    "carries out the jit (just in time requirements between options"
    def build_actions(requires):
        "action are hide, show, enable, disable..."
        trigger_actions = {}
        for require in requires:
            action = require[2]
            trigger_actions.setdefault(action, []).append(require)
        return trigger_actions
    #for symlink
    if hasattr(opt, '_requires') and opt._requires is not None:
        # filters the callbacks
        setting = config.cfgimpl_get_settings()
        trigger_actions = build_actions(opt._requires)
        optpath = config.cfgimpl_get_context().cfgimpl_get_description().get_path_by_opt(opt)
        for requires in trigger_actions.values():
            matches = False
            for require in requires:
                if len(require) == 3:
                    path, expected, action = require
                    inverse = False
                elif len(require) == 4:
                    path, expected, action, inverse = require
                if path == optpath or path.startswith(optpath + '.'):
                    raise RequirementRecursionError(_("malformed requirements "
                                                    "imbrication detected for option: '{0}' "
                                                    "with requirement on: '{1}'").format(optpath, path))
                try:
                    value = config.cfgimpl_get_context()._getattr(path, force_permissive=True)
                except PropertiesOptionError, err:
                    properties = err.proptype
                    #FIXME: AttributeError or PropertiesOptionError ?
                    raise AttributeError(_("option '{0}' has requirement's property error: "
                                         "{1} {2}").format(opt._name, path, properties))
                except AttributeError:
                    raise AttributeError(_("required option not found: "
                                         "{0}").format(path))
                if value == expected:
                    if inverse:
                        setting.del_property(action, opt, False)
                    else:
                        setting.add_property(action, opt, False)
                    matches = True
                    #FIXME optimisation : fait un double break non ? voire un return
            # no requirement has been triggered, then just reverse the action
            if not matches:
                if inverse:
                    setting.add_property(action, opt, False)
                else:
                    setting.del_property(action, opt, False)
